# Copyright (C) 2010-2012 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Exceptions"""

class SamlError(Exception):
  """A fatal SAML error. We are unable to send a request/response."""
